/*******************************************************************************
 * Copyright (c) 2004, 2005 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *******************************************************************************/

package org.eclipse.birt.data.engine.aggregation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.birt.core.framework.FrameworkException;
import org.eclipse.birt.core.framework.IConfigurationElement;
import org.eclipse.birt.core.framework.IExtension;
import org.eclipse.birt.core.framework.IExtensionPoint;
import org.eclipse.birt.core.framework.IExtensionRegistry;
import org.eclipse.birt.core.framework.Platform;
import org.eclipse.birt.data.engine.api.aggregation.IAggregation;
import org.eclipse.birt.data.engine.api.aggregation.IAggregationFactory;
import org.eclipse.birt.data.engine.api.aggregation.IAggregationInfo;
import org.eclipse.birt.data.engine.api.aggregation.IBuildInAggregation;
import org.eclipse.birt.data.engine.core.DataException;
import org.eclipse.birt.data.engine.i18n.ResourceConstants;


/**
 * The instances of this class manage all aggregations, including BuiltInAggregation (as defined
 * in BuiltInAggregationFactory) and customized aggregation (as defined in plugins).
 */

public class AggregationFactory implements IAggregationFactory 
{
	// model aggregation function names
	private static final String MODEL_RANK = "rank";
	private static final String MODEL_PERCENT_SUM = "percent-sum";
	private static final String MODEL_QUARTILE = "quartile";
	private static final String MODEL_PERCENTILE = "percentile";
	private static final String MODEL_PERCENT_RANK = "percent-rank";
	private static final String MODEL_IS_BOTTOM_N_PERCENT = "is-bottom-n-percent";
	private static final String MODEL_IS_TOP_N_PERCENT = "is-top-n-percent";
	private static final String MODEL_IS_BOTTOM_N = "is-bottom-n";
	private static final String MODEL_IS_TOP_N = "is-top-n";
	private static final String MODEL_RUNNING_COUNT = "running-count";
	private static final String MODEL_COUNT_DISTINCT = "count-distinct";
	private static final String MODEL_RUNNING_NPV = "running-npv";
	private static final String MODEL_NPV = "npv";
	private static final String MODEL_MIRR = "mirr";
	private static final String MODEL_IRR = "irr";
	private static final String MODEL_RUNNING_SUM = "running-sum";
	private static final String MODEL_VARIANCE = "variance";
	private static final String MODEL_MEDIAN = "median";
	private static final String MODEL_MOVING_AVE = "moving-ave";
	private static final String MODEL_MODE = "mode";
	private static final String MODEL_LAST = "last";
	private static final String MODEL_FIRST = "first";
	private static final String MODEL_STDDEV = "stddev";
	private static final String MODEL_WEIGHTED_AVG = "weighted-avg";
	private static final String MODEL_AVERAGE = "average";
	private static final String MODEL_MAX = "max";
	private static final String MODEL_MIN = "min";
	private static final String MODEL_COUNT = "count";
	private static final String MODEL_SUM = "sum";
	
	private static final String ATTRIBUTE_AGGREGATION_CLASS = "aggregationClass";
	private static final String ATTRIBUTE_AGGREGATION_NAME = "name";
	private static final String EXTENSION_POINT = "org.eclipse.birt.data.aggregation";
	private static final String ELEMENT_AGGREGATIONS = "Aggregations";
	private static final String ELEMENT_AGGREGATION = "Aggregation";
	
	private static final String ELEMENT_UIINFO = "UIInfo";
	private static final String ATTRIBUTE_PARAMETER_META_INFO = "parameterMetaInfo";
	private static final String ATTRIBUTE_TEXT_DATA = "textData";
	
	private static IAggregationFactory instance = null;
	private static Map aggregations = null;
	private static Map aggrAdapterMap = null;
	private static Map aggrInfoMap = null;
	
	/**
	 * allowed aggregation function names in x-tab
	 */
	private static String[] xTabAggrNames = new String[]{
			IBuildInAggregation.TOTAL_SUM_FUNC,
			IBuildInAggregation.TOTAL_AVE_FUNC,
			IBuildInAggregation.TOTAL_MAX_FUNC,
			IBuildInAggregation.TOTAL_MIN_FUNC,
			IBuildInAggregation.TOTAL_FIRST_FUNC,
			IBuildInAggregation.TOTAL_LAST_FUNC,
			IBuildInAggregation.TOTAL_COUNT_FUNC,
			IBuildInAggregation.TOTAL_COUNTDISTINCT_FUNC
	};
	
	/**
	 * allowed aggregation function names in cube measure.
	 */
	private static String[] measureAggrNames = new String[]{
			IBuildInAggregation.TOTAL_SUM_FUNC,
			IBuildInAggregation.TOTAL_MAX_FUNC,
			IBuildInAggregation.TOTAL_MIN_FUNC,
			IBuildInAggregation.TOTAL_FIRST_FUNC,
			IBuildInAggregation.TOTAL_LAST_FUNC,
			IBuildInAggregation.TOTAL_COUNT_FUNC,
			IBuildInAggregation.TOTAL_COUNTDISTINCT_FUNC
	};
	
	private static List tabularAggrNames = new ArrayList( );


	/**
	 * Return a shared instance of AggregationFactory.
	 * @return
	 * @throws DataException
	 */
	public static IAggregationFactory getInstance( ) throws DataException
	{
		if ( instance == null )
		{
			synchronized ( AggregationFactory.class )
			{
				if ( instance == null )
					instance = new AggregationFactory( );
			}
		}
		
		return instance;
	}
	
	/**
	 * Destroy shared instance of AggregationFactory.
	 * 
	 */
	public static void destroyInstance( )
	{
		instance = null;
		aggregations = null;
		aggrAdapterMap = null;
		aggrInfoMap = null;
	}

	
	private AggregationFactory( ) throws DataException
	{
		populateAggregations( );
	}

	/**
	 * Populate the aggregations. Put all aggregations to the HashMap "aggregations" 
	 * @throws DataException
	 */
	private void populateAggregations( ) throws DataException
	{
		aggregations = new HashMap( );
		aggrAdapterMap = new HashMap( );
		aggrInfoMap = new HashMap( );
		populateAggrAdapterMap( );
		populateBuiltInAggregations( );
		populateExtendedAggregations( );
		populateAggrInfo( );
	}

	/**
	 * populate the build-in aggregation's information to the HashMap
	 * "aggrInfoMap".
	 */
	private void populateAggrInfo( )
	{
		// sum
		AggregationInfo aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_SUM_FUNC,
				ResourceConstants.TOTAL_FUNC_SUM );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_SUM_FUNC, aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_SUM_FUNC);
		// running-sum
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_RUNNINGSUM_FUNC,
				ResourceConstants.TOTAL_FUNC_RUNNINGSUM );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_RUNNINGSUM_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_RUNNINGSUM_FUNC);
		// count
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_COUNT_FUNC,
				ResourceConstants.TOTAL_FUNC_COUNT );
		aggrInfo.setNeedDataField( false );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_COUNT_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_COUNT_FUNC);
		// running-count
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_RUNNINGCOUNT_FUNC,
				ResourceConstants.TOTAL_FUNC_RUNNINGCOUNT );
		aggrInfo.setNeedDataField( false );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_RUNNINGCOUNT_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_RUNNINGCOUNT_FUNC);
		// count-distinct
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_COUNTDISTINCT_FUNC,
				ResourceConstants.TOTAL_FUNC_COUNTDISTINCT );
		aggrInfo.setNeedDataField( true );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_COUNTDISTINCT_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_COUNTDISTINCT_FUNC);
		// max
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_MAX_FUNC,
				ResourceConstants.TOTAL_FUNC_MAX );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_MAX_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_MAX_FUNC);
		// min
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_MIN_FUNC,
				ResourceConstants.TOTAL_FUNC_MIN );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_MIN_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_MIN_FUNC);
		// average
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_AVE_FUNC,
				ResourceConstants.TOTAL_FUNC_AVE );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_AVE_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_AVE_FUNC);
		// weighted-avg
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_WEIGHTEDAVE_FUNC,
				ResourceConstants.TOTAL_FUNC_WEIGHTEDAVE );
		ParameterInfo paramInfo = new ParameterInfo( "weight",
				ResourceConstants.TOTAL_PARAM_WEIGHTEDAVE_WEIGHT );
		aggrInfo.addParameter( paramInfo );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_WEIGHTEDAVE_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_WEIGHTEDAVE_FUNC);
		// moving-ave
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_MOVINGAVE_FUNC,
				ResourceConstants.TOTAL_FUNC_MOVINGAVE );
		paramInfo = new ParameterInfo( "window",
				ResourceConstants.TOTAL_PARAM_MOVINGAVE_WINDOW );
		aggrInfo.addParameter( paramInfo );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_MOVINGAVE_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_MOVINGAVE_FUNC);
		// median
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_MEDIAN_FUNC,
				ResourceConstants.TOTAL_FUNC_MEDIAN );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_MEDIAN_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_MEDIAN_FUNC);
		// mode
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_MODE_FUNC,
				ResourceConstants.TOTAL_FUNC_MODE );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_MODE_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_MODE_FUNC);
		// stddev
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_STDDEV_FUNC,
				ResourceConstants.TOTAL_FUNC_STDDEV );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_STDDEV_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_STDDEV_FUNC);
		// variance
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_VARIANCE_FUNC,
				ResourceConstants.TOTAL_FUNC_VARIANCE );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_VARIANCE_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_VARIANCE_FUNC);
		// first
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_FIRST_FUNC,
				ResourceConstants.TOTAL_FUNC_FIRST );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_FIRST_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_FIRST_FUNC);
		// last
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_LAST_FUNC,
				ResourceConstants.TOTAL_FUNC_LAST );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_LAST_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_LAST_FUNC);
		// rank
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_RANK_FUNC,
				ResourceConstants.TOTAL_FUNC_RANK );
		paramInfo = new ParameterInfo( "ascending",
				ResourceConstants.TOTAL_PARAM_RANK_ASCENDING );
		aggrInfo.addParameter( paramInfo );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_RANK_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_RANK_FUNC);
		// percent-rank
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_PERCENT_RANK_FUNC,
				ResourceConstants.TOTAL_FUNC_PERCENTRANK );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_PERCENT_RANK_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_PERCENT_RANK_FUNC);
		// percentile
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_PERCENTILE_FUNC,
				ResourceConstants.TOTAL_FUNC_PERCENTILE );
		paramInfo = new ParameterInfo( "pct",
				ResourceConstants.TOTAL_PARAM_PERCENTILE_PCT );
		aggrInfo.addParameter( paramInfo );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_PERCENTILE_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_PERCENTILE_FUNC);
		// quartile
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_QUARTILE_FUNC,
				ResourceConstants.TOTAL_FUNC_QUARTILE );
		paramInfo = new ParameterInfo( IBuildInAggregation.TOTAL_QUARTILE_FUNC,
				ResourceConstants.TOTAL_PARAM_QUARTILE_QUART );
		aggrInfo.addParameter( paramInfo );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_QUARTILE_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_QUARTILE_FUNC);
		// percent-sum
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_PERCENTSUM_FUNC,
				ResourceConstants.TOTAL_FUNC_PERCENTSUM );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_PERCENTSUM_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_PERCENTSUM_FUNC);
		// top-n
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_TOP_N_FUNC,
				ResourceConstants.TOTAL_FUNC_ISTOPN );
		paramInfo = new ParameterInfo( IBuildInAggregation.TOTAL_TOP_N_FUNC,
				ResourceConstants.TOTAL_PARAM_ISTOPN_N );
		aggrInfo.addParameter( paramInfo );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_TOP_N_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_TOP_N_FUNC);
		// top-percent
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_TOP_PERCENT_FUNC,
				ResourceConstants.TOTAL_FUNC_ISTOPNPERCENT );
		paramInfo = new ParameterInfo( IBuildInAggregation.TOTAL_TOP_PERCENT_FUNC,
				ResourceConstants.TOTAL_PARAM_ISTOPNPERCENT_PCT );
		aggrInfo.addParameter( paramInfo );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_TOP_PERCENT_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_TOP_PERCENT_FUNC);
		// bottom-n
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_BOTTOM_N_FUNC,
				ResourceConstants.TOTAL_FUNC_ISBOTTOMN );
		paramInfo = new ParameterInfo( IBuildInAggregation.TOTAL_BOTTOM_N_FUNC,
				ResourceConstants.TOTAL_PARAM_ISBOTTOMN_N );
		aggrInfo.addParameter( paramInfo );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_BOTTOM_N_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_BOTTOM_N_FUNC);
		// bottom-percent
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_BOTTOM_PERCENT_FUNC,
				ResourceConstants.TOTAL_FUNC_ISBOTTOMNPERCENT );
		paramInfo = new ParameterInfo( IBuildInAggregation.TOTAL_BOTTOM_PERCENT_FUNC,
				ResourceConstants.TOTAL_PARAM_ISBOTTOMNPERCENT_PCT );
		aggrInfo.addParameter( paramInfo );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_BOTTOM_PERCENT_FUNC,aggrInfo );
		// irr
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_IRR_FUNC,
				ResourceConstants.TOTAL_FUNC_IRR );
		paramInfo = new ParameterInfo( "startingGuess",
				ResourceConstants.TOTAL_PARAM_IRR_RATE );
		paramInfo.setOptional( true );
		aggrInfo.addParameter( paramInfo );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_IRR_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_IRR_FUNC);
		// mirr
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_MIRR_FUNC,
				ResourceConstants.TOTAL_FUNC_MIRR );
		paramInfo = new ParameterInfo( "financeRate",
				ResourceConstants.TOTAL_PARAM_MIRR_FINANCE_RATE );
		paramInfo.setOptional( true );
		aggrInfo.addParameter( paramInfo );
		paramInfo = new ParameterInfo( "reinvestmentRate",
				ResourceConstants.TOTAL_PARAM_MIRR_REINVESTMENT_RATE );
		paramInfo.setOptional( true );
		aggrInfo.addParameter( paramInfo );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_MIRR_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_MIRR_FUNC);
		// npv
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_NPV_FUNC,
				ResourceConstants.TOTAL_FUNC_NPV );
		paramInfo = new ParameterInfo( "rate",
				ResourceConstants.TOTAL_PARAM_NPV_RATE );
		paramInfo.setOptional( true );
		aggrInfo.addParameter( paramInfo );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_NPV_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_NPV_FUNC);
		// running-npv
		aggrInfo = new AggregationInfo( IBuildInAggregation.TOTAL_RUNNINGNPV_FUNC,
				ResourceConstants.TOTAL_FUNC_RUNNINGNPV );
		paramInfo = new ParameterInfo( "rate",
				ResourceConstants.TOTAL_PARAM_RUNNINGNPV_RATE );
		paramInfo.setOptional( true );
		aggrInfo.addParameter( paramInfo );
		aggrInfoMap.put( IBuildInAggregation.TOTAL_RUNNINGNPV_FUNC,aggrInfo );
		tabularAggrNames.add( IBuildInAggregation.TOTAL_RUNNINGNPV_FUNC);
	}


	/**
	 * Populate model aggregation names to build in aggregation names.
	 * Note: This adapter should be synchronized with DataAdapterUtil in Model.
	 */
	private static void populateAggrAdapterMap( )
	{
		aggrAdapterMap.put( MODEL_SUM, IBuildInAggregation.TOTAL_SUM_FUNC );
		aggrAdapterMap.put( MODEL_COUNT, IBuildInAggregation.TOTAL_COUNT_FUNC );
		aggrAdapterMap.put( MODEL_MIN, IBuildInAggregation.TOTAL_MIN_FUNC );
		aggrAdapterMap.put( MODEL_MAX, IBuildInAggregation.TOTAL_MAX_FUNC );
		aggrAdapterMap.put( MODEL_AVERAGE, IBuildInAggregation.TOTAL_AVE_FUNC );
		aggrAdapterMap.put( MODEL_WEIGHTED_AVG, IBuildInAggregation.TOTAL_WEIGHTEDAVE_FUNC );
		aggrAdapterMap.put( MODEL_STDDEV, IBuildInAggregation.TOTAL_STDDEV_FUNC );
		aggrAdapterMap.put( MODEL_FIRST, IBuildInAggregation.TOTAL_FIRST_FUNC );
		aggrAdapterMap.put( MODEL_LAST, IBuildInAggregation.TOTAL_LAST_FUNC );
		aggrAdapterMap.put( MODEL_MODE, IBuildInAggregation.TOTAL_MODE_FUNC );
		aggrAdapterMap.put( MODEL_MOVING_AVE, IBuildInAggregation.TOTAL_MOVINGAVE_FUNC );
		aggrAdapterMap.put( MODEL_MEDIAN, IBuildInAggregation.TOTAL_MEDIAN_FUNC );
		aggrAdapterMap.put( MODEL_VARIANCE, IBuildInAggregation.TOTAL_VARIANCE_FUNC );
		aggrAdapterMap.put( MODEL_RUNNING_SUM, IBuildInAggregation.TOTAL_RUNNINGSUM_FUNC );
		aggrAdapterMap.put( MODEL_IRR, IBuildInAggregation.TOTAL_IRR_FUNC );
		aggrAdapterMap.put( MODEL_MIRR, IBuildInAggregation.TOTAL_MIRR_FUNC );
		aggrAdapterMap.put( MODEL_NPV, IBuildInAggregation.TOTAL_NPV_FUNC );
		aggrAdapterMap.put( MODEL_RUNNING_NPV, IBuildInAggregation.TOTAL_RUNNINGNPV_FUNC );
		aggrAdapterMap.put( MODEL_COUNT_DISTINCT, IBuildInAggregation.TOTAL_COUNTDISTINCT_FUNC );
		aggrAdapterMap.put( MODEL_RUNNING_COUNT, IBuildInAggregation.TOTAL_RUNNINGCOUNT_FUNC );
		aggrAdapterMap.put( MODEL_IS_TOP_N, IBuildInAggregation.TOTAL_TOP_N_FUNC );
		aggrAdapterMap.put( MODEL_IS_BOTTOM_N, IBuildInAggregation.TOTAL_BOTTOM_N_FUNC );
		aggrAdapterMap.put( MODEL_IS_TOP_N_PERCENT, IBuildInAggregation.TOTAL_TOP_PERCENT_FUNC );
		aggrAdapterMap.put( MODEL_IS_BOTTOM_N_PERCENT, IBuildInAggregation.TOTAL_BOTTOM_PERCENT_FUNC );
		aggrAdapterMap.put( MODEL_PERCENT_RANK, IBuildInAggregation.TOTAL_PERCENT_RANK_FUNC );
		aggrAdapterMap.put( MODEL_PERCENTILE, IBuildInAggregation.TOTAL_PERCENTILE_FUNC );
		aggrAdapterMap.put( MODEL_QUARTILE, IBuildInAggregation.TOTAL_QUARTILE_FUNC );
		aggrAdapterMap.put( MODEL_PERCENT_SUM, IBuildInAggregation.TOTAL_PERCENTSUM_FUNC );
		aggrAdapterMap.put( MODEL_RANK, IBuildInAggregation.TOTAL_RANK_FUNC );
	}
	
	/**
	 * populate the built-in Aggregations
	 */
	private void populateBuiltInAggregations( )
	{
		String[] builtInAggrNames = BuiltInAggregationFactory.getInstance( )
				.getAggregationNames( );
		for ( int i = 0; i < builtInAggrNames.length; i++ )
		{
			aggregations.put( builtInAggrNames[i],
					BuiltInAggregationFactory.getInstance( )
							.getAggregation( builtInAggrNames[i] ) );
		}
	}

	/**
	 * Load aggregation extensions from runtime environment.
	 * @throws DataException 
	 * @throws FrameworkException 
	 */
	private void populateExtendedAggregations( ) throws DataException 
	{
		IExtensionRegistry extReg = Platform.getExtensionRegistry( );

		IExtensionPoint extPoint = extReg.getExtensionPoint( EXTENSION_POINT );

		if ( extPoint == null )
			return;

		IExtension[] exts = extPoint.getExtensions( );
		if ( exts == null )
			return;

		for ( int e = 0; e < exts.length; e++ )
		{
			IConfigurationElement[] configElems = exts[e].getConfigurationElements( );
			if ( configElems == null )
				continue;

			for ( int i = 0; i < configElems.length; i++ )
			{
				if ( configElems[i].getName( ).equals( ELEMENT_AGGREGATIONS ) )
				{
					IConfigurationElement[] subElems = configElems[i].getChildren( ELEMENT_AGGREGATION );
					for ( int j = 0; j < subElems.length; j++ )
					{

						String name = subElems[j].getAttribute( ATTRIBUTE_AGGREGATION_NAME )
								.toUpperCase( )
								.trim( );
						try
						{
							IAggregation aggrFunc = (IAggregation) subElems[j].createExecutableExtension( ATTRIBUTE_AGGREGATION_CLASS );
							populateExtendedAggrInfo( name,
									aggrFunc,
									subElems[j] );
							if ( aggregations.put( name, aggrFunc ) != null )
								throw new DataException( ResourceConstants.DUPLICATE_AGGREGATION_NAME,
										name );
						}
						catch ( FrameworkException exception )
						{
							// TODO: log this exception or provide public
							// interface for the user to get uninstantiated
							// function names
						}
					}
				}
			}
		}
	}
	
	/**
	 * populate the extended extensions information.
	 * @param name
	 * @param aggrFunc
	 * @param elem
	 */
	private void populateExtendedAggrInfo( String name, IAggregation aggrFunc,
			IConfigurationElement elem )
	{
		IConfigurationElement[] uiInfo = elem.getChildren( ELEMENT_UIINFO );
		assert ( uiInfo != null && uiInfo.length == 1 );
		String paramInfo = uiInfo[0].getAttribute( ATTRIBUTE_PARAMETER_META_INFO );
		String textInfo = uiInfo[0].getAttribute( ATTRIBUTE_TEXT_DATA );
		AggregationInfo aggrInfo = new AggregationInfo( name );
		aggrInfo.setDisplayName( textInfo );
		String[] paramInfos = paramInfo.split( "," );
		boolean[] paramFlags = aggrFunc.getParameterDefn( );
		if ( paramInfos != null && paramInfos.length > 0 )
		{
			for ( int k = 0; k < paramInfos.length; k++ )
			{
				final String s = paramInfos[k].trim( );
				int index = s.indexOf( ' ' );
				String paramName = null;
				if ( index > 0 )
				{
					paramName = s.substring( index + 1 ).trim( );
				}
				else
				{
					paramName = paramInfos[k];
				}
				if ( k + 1 >= paramFlags.length )
				{
					break;
				}
				ParameterInfo argInfo = new ParameterInfo( paramName );
				argInfo.setOptional( !paramFlags[k + 1] );
				aggrInfo.addParameter( argInfo );
			}
		}
		aggrInfoMap.put( name, aggrInfo );
		tabularAggrNames.add( name );
	}

	
	/* (non-Javadoc)
	 * @see org.eclipse.birt.data.engine.api.aggregation.IAggregationFactory#getAggrInfoList(int)
	 */
	public List getAggrInfoList( int type )
	{
		switch ( type )
		{
			case AGGR_TABULAR :
				return getResult( tabularAggrNames.toArray( ) );
			case AGGR_XTAB :
				return getResult( xTabAggrNames );
			case AGGR_MEASURE :
				return getResult( measureAggrNames );
		}
		return new ArrayList( );
	}

	/**
	 * 
	 * @param names
	 * @return
	 */
	private List getResult( Object[] names )
	{
		List list = new ArrayList( );
		for ( int i = 0; i < names.length; i++ )
		{
			Object aggrInfo = aggrInfoMap.get( names[i] );
			if ( aggrInfo != null )
			{
				list.add( aggrInfo );
			}
		}
		return list;
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.birt.data.engine.api.aggregation.IAggregationFactory#getAggregation(java.lang.String)
	 */
	public IAggregation getAggregation( String name )
	{
		assert name != null;
		// try to fetch the aggregation as build-in aggregation name
		IAggregation aggr = (IAggregation) aggregations.get( name.toUpperCase( ) );
		// if fails, it may be an aggregation name used in model and will be adapted to
		// build-in aggregation name
		if ( aggr == null )
		{
			String buildInAggrName = (String) aggrAdapterMap.get( name );
			aggr = (IAggregation) aggregations.get( buildInAggrName );
		}
		return aggr;
	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.birt.data.engine.api.aggregation.IAggregationFactory#getAggrInfo(java.lang.String)
	 */
	public IAggregationInfo getAggrInfo( String aggrName )
	{
		return (IAggregationInfo) aggrInfoMap.get( aggrName );
	}
}
